import { compose } from 'redux';

const prefix = '@@redux-ops/';
const START = `${prefix}START`;
const UPDATE = `${prefix}UPDATE`;
const DELETE = `${prefix}DELETE`;
const RESET = `${prefix}RESET`;
const BLUEPRINT = `${prefix}BLUEPRINT`;

var action_types = /*#__PURE__*/Object.freeze({
  __proto__: null,
  prefix: prefix,
  START: START,
  UPDATE: UPDATE,
  DELETE: DELETE,
  RESET: RESET,
  BLUEPRINT: BLUEPRINT
});

const defaultState = {};
const updateOperation = (state, action) => {
    const { payload } = action;
    return Object.assign(Object.assign({}, state), { [payload.id]: Object.assign({}, payload) });
};
const deleteOperation = (state, action) => {
    const newState = Object.assign({}, state);
    delete newState[action.payload.id];
    return newState;
};
const resetOperations = () => (Object.assign({}, defaultState));
const handlers = {
    [START]: updateOperation,
    [UPDATE]: updateOperation,
    [DELETE]: deleteOperation,
    [RESET]: resetOperations,
};
const reducer = (state = defaultState, action) => {
    const handler = handlers[action.type];
    return handler ? handler(state, action) : state;
};

var OpStatus;
(function (OpStatus) {
    OpStatus["Started"] = "started";
    OpStatus["Success"] = "success";
    OpStatus["Error"] = "error";
})(OpStatus || (OpStatus = {}));

function createOperation(id, status = OpStatus.Started, data) {
    return { id, status, data };
}
function getOpById(ops, id) {
    return ops[id];
}
function getOpByIds(ops, ids) {
    return ids.reduce((acc, id) => {
        const op = ops[id];
        op && (acc[id] = op);
        return acc;
    }, {});
}
function searchOps(ops, query) {
    return Object.keys(ops).reduce((acc, id) => {
        if (query instanceof RegExp ? !!String(id).match(query) : String(id).indexOf(query) > -1) {
            acc[id] = ops[id];
        }
        return acc;
    }, {});
}

var utils = /*#__PURE__*/Object.freeze({
  __proto__: null,
  createOperation: createOperation,
  getOpById: getOpById,
  getOpByIds: getOpByIds,
  searchOps: searchOps
});

const startOperation = (id, data) => ({
    type: START,
    payload: createOperation(id, OpStatus.Started, data),
});
const updateOperation$1 = (id, status, data) => ({
    type: UPDATE,
    payload: createOperation(id, status, data),
});
const deleteOperation$1 = id => ({
    type: DELETE,
    payload: { id },
});
const resetOperations$1 = () => ({
    type: RESET,
});

var actions = /*#__PURE__*/Object.freeze({
  __proto__: null,
  startOperation: startOperation,
  updateOperation: updateOperation$1,
  deleteOperation: deleteOperation$1,
  resetOperations: resetOperations$1
});

let reducerName = 'ops';
function setReducerName(name) {
    reducerName = name;
}
function getOpsState(state) {
    const subState = state[reducerName];
    if (subState === undefined) {
        throw new Error(`No reducer found with name ${reducerName}. Try to set it via setReducerName().`);
    }
    return subState;
}
function getOps(state) {
    return getOpsState(state);
}
function getOpById$1(state, id) {
    return getOpById(getOps(state), id);
}
function getOpByIds$1(state, ids) {
    return getOpByIds(getOps(state), ids);
}
function searchOps$1(state, query) {
    return searchOps(getOps(state), query);
}

var selectors = /*#__PURE__*/Object.freeze({
  __proto__: null,
  setReducerName: setReducerName,
  getOpsState: getOpsState,
  getOps: getOps,
  getOpById: getOpById$1,
  getOpByIds: getOpByIds$1,
  searchOps: searchOps$1
});

var OpsBlueprintActionKey;
(function (OpsBlueprintActionKey) {
    OpsBlueprintActionKey["Start"] = "start";
    OpsBlueprintActionKey["Success"] = "success";
    OpsBlueprintActionKey["Error"] = "error";
    OpsBlueprintActionKey["Delete"] = "delete";
})(OpsBlueprintActionKey || (OpsBlueprintActionKey = {}));

let uniqueCounter = 0;
function createBlueprintAction(op, action) {
    return {
        type: BLUEPRINT,
        [prefix]: { op, action },
    };
}
function composeBlueprint(blueprint, composer) {
    return Object.assign(Object.assign({}, blueprint), Object.keys(OpsBlueprintActionKey).reduce((acc, value) => {
        const key = OpsBlueprintActionKey[value];
        acc[key] = compose(composer, blueprint[key]);
        return acc;
    }, {}));
}
function composeActions(blueprintActionCreator, actionCreator) {
    return actionCreator
        ? compose((action, data) => blueprintActionCreator(data, action), actionCreator)
        : data => blueprintActionCreator(data);
}
function createStartAction(id) {
    return (data, action) => createBlueprintAction(startOperation(id, data), action);
}
function createUpdateAction(id, opStatus) {
    return (data, action) => createBlueprintAction(updateOperation$1(id, opStatus, data), action);
}
function createDeleteAction(id) {
    return (data, action) => createBlueprintAction(deleteOperation$1(id), action);
}
function createBlueprint(id, composers = {}) {
    return Object.assign({ id, start: composeActions(createStartAction(id), composers.start), success: composeActions(createUpdateAction(id, OpStatus.Success), composers.success), error: composeActions(createUpdateAction(id, OpStatus.Error), composers.error), delete: composeActions(createDeleteAction(id), composers.delete) }, createBlueprintActionTypes(id));
}
function createBlueprintActionTypes(id) {
    return Object.keys(OpsBlueprintActionKey).reduce((acc, value) => {
        const key = value.toUpperCase();
        acc[key] = `${String(id)}_${key}`;
        return acc;
    }, {});
}
function opsUnique(blueprint, uniqueValue) {
    if (blueprint[prefix]) {
        const opsData = blueprint[prefix];
        const uniqueId = generateUniqueId(opsData.op.payload.id, uniqueValue);
        if (opsData.action) {
            opsData.action[prefix] = { uniqueId };
        }
        blueprint[prefix].uniqueId = uniqueId;
        return blueprint;
    }
    const uniqueId = generateUniqueId(blueprint.id, uniqueValue);
    return composeBlueprint(blueprint, (action) => opsUnique(action, uniqueId));
}
function generateUniqueId(id, uniqueValue) {
    return uniqueValue
        ? typeof uniqueValue === 'object'
            ? getUniqueId(uniqueValue)
            : uniqueValue
        : `${prefix}${id}_${++uniqueCounter}`;
}
function opsBroadcast(blueprint) {
    if (blueprint[prefix]) {
        blueprint[prefix].broadcast = true;
        return blueprint;
    }
    return composeBlueprint(blueprint, opsBroadcast);
}
function getUniqueId(action) {
    const opsData = action[prefix];
    return opsData && opsData.uniqueId;
}

function createOpsMiddleware() {
    return store => next => action => {
        if (action[prefix]) {
            const blueprintAction = action[prefix];
            let opAction = blueprintAction.op;
            const opActionPayload = opAction.payload;
            const originalAction = blueprintAction.action;
            const opStatus = 'status' in opActionPayload ? opActionPayload.status : undefined;
            let broadcastAction = undefined;
            if (blueprintAction.broadcast) {
                let broadcastActionType = opAction.type;
                if (opStatus && broadcastActionType === UPDATE) {
                    broadcastActionType = opStatus.toUpperCase();
                }
                else {
                    broadcastActionType = broadcastActionType.replace(prefix, '');
                }
                broadcastAction = {
                    type: `${opActionPayload.id}_${broadcastActionType}`,
                    payload: Object.assign({}, opActionPayload),
                };
            }
            if (opStatus) {
                if (blueprintAction.broadcast) {
                    opAction = Object.assign(Object.assign({}, opAction), { payload: Object.assign(Object.assign({}, opAction.payload), { data: undefined }) });
                }
            }
            if (blueprintAction.uniqueId) {
                const { uniqueId } = blueprintAction;
                opAction = Object.assign(Object.assign({}, opAction), { payload: Object.assign(Object.assign({}, opActionPayload), { id: uniqueId }) });
                if (broadcastAction) {
                    broadcastAction[prefix] = { uniqueId };
                }
            }
            switch (opStatus) {
                case OpStatus.Started: {
                    next(opAction);
                    originalAction && next(originalAction);
                    broadcastAction && next(broadcastAction);
                    return;
                }
                default: {
                    originalAction && next(originalAction);
                    broadcastAction && next(broadcastAction);
                    return next(opAction);
                }
            }
        }
        return next(action);
    };
}
var middleware = createOpsMiddleware();

export default reducer;
export { OpStatus, OpsBlueprintActionKey, action_types as actionTypes, actions, composeBlueprint, createBlueprint, createBlueprintAction, defaultState, getUniqueId, opsBroadcast, middleware as opsMiddleware, opsUnique, selectors, utils };
